@@ -399,7 +399,7 @@ PXENV_EXIT_t pxenv_undi_transmit ( t_PXENV_UNDI_TRANSMIT *undi_transmit ) {
 
 /* PXENV_UNDI_SET_MCAST_ADDRESS
  *
- * Status: stub
+ * Status: stub (no PXE multicast support)
  */
 PXENV_EXIT_t pxenv_undi_set_mcast_address ( t_PXENV_UNDI_SET_MCAST_ADDRESS
 					    *undi_set_mcast_address ) {
@@ -436,7 +436,8 @@ PXENV_EXIT_t pxenv_undi_set_station_address ( t_PXENV_UNDI_SET_STATION_ADDRESS
 
 /* PXENV_UNDI_SET_PACKET_FILTER
  *
- * Status: stub
+ * Status: won't implement (would require driver API changes for no
+ * real benefit)
  */
 PXENV_EXIT_t pxenv_undi_set_packet_filter ( t_PXENV_UNDI_SET_PACKET_FILTER
 					    *undi_set_packet_filter ) {
@@ -482,7 +483,8 @@ PXENV_EXIT_t pxenv_undi_get_information ( t_PXENV_UNDI_GET_INFORMATION
 
 /* PXENV_UNDI_GET_STATISTICS
  *
- * Status: stub
+ * Status: won't implement (would require driver API changes for no
+ * real benefit)
  */
 PXENV_EXIT_t pxenv_undi_get_statistics ( t_PXENV_UNDI_GET_STATISTICS
 					 *undi_get_statistics ) {
@@ -494,7 +496,8 @@ PXENV_EXIT_t pxenv_undi_get_statistics ( t_PXENV_UNDI_GET_STATISTICS
 
 /* PXENV_UNDI_CLEAR_STATISTICS
  *
- * Status: stub
+ * Status: won't implement (would require driver API changes for no
+ * real benefit)
  */
 PXENV_EXIT_t pxenv_undi_clear_statistics ( t_PXENV_UNDI_CLEAR_STATISTICS
 					   *undi_clear_statistics ) {
@@ -506,7 +509,8 @@ PXENV_EXIT_t pxenv_undi_clear_statistics ( t_PXENV_UNDI_CLEAR_STATISTICS
 
 /* PXENV_UNDI_INITIATE_DIAGS
  *
- * Status: stub
+ * Status: won't implement (would require driver API changes for no
+ * real benefit)
  */
 PXENV_EXIT_t pxenv_undi_initiate_diags ( t_PXENV_UNDI_INITIATE_DIAGS
 					 *undi_initiate_diags ) {
@@ -518,19 +522,21 @@ PXENV_EXIT_t pxenv_undi_initiate_diags ( t_PXENV_UNDI_INITIATE_DIAGS
 
 /* PXENV_UNDI_FORCE_INTERRUPT
  *
- * Status: stub
+ * Status: working
  */
 PXENV_EXIT_t pxenv_undi_force_interrupt ( t_PXENV_UNDI_FORCE_INTERRUPT
 					  *undi_force_interrupt ) {
 	DBG ( "PXENV_UNDI_FORCE_INTERRUPT" );
-	/* ENSURE_READY ( undi_force_interrupt ); */
-	undi_force_interrupt->Status = PXENV_STATUS_UNSUPPORTED;
-	return PXENV_EXIT_FAILURE;
+	ENSURE_READY ( undi_force_interrupt );
+
+	eth_irq ( FORCE );
+	undi_force_interrupt->Status = PXENV_STATUS_SUCCESS;
+	return PXENV_EXIT_SUCCESS;
 }
 
 /* PXENV_UNDI_GET_MCAST_ADDRESS
  *
- * Status: stub
+ * Status: stub (no PXE multicast support)
  */
 PXENV_EXIT_t pxenv_undi_get_mcast_address ( t_PXENV_UNDI_GET_MCAST_ADDRESS
 					    *undi_get_mcast_address ) {
@@ -613,7 +619,7 @@ PXENV_EXIT_t pxenv_undi_get_iface_info ( t_PXENV_UNDI_GET_IFACE_INFO
 
 /* PXENV_UNDI_ISR
  *
- * Status: working when hooked to the timer interrupt
+ * Status: working
  */
 PXENV_EXIT_t pxenv_undi_isr ( t_PXENV_UNDI_ISR *undi_isr ) {
 	media_header_t *media_header = (media_header_t*)nic.packet;
@@ -688,18 +694,23 @@ PXENV_EXIT_t pxenv_undi_isr ( t_PXENV_UNDI_ISR *undi_isr ) {
 			} else {
 				undi_isr->PktType = XMT_DESTADDR;
 			}
+			break;
 		} else {
-			DBG ( " DONE" );
-			/* Re-enable interrupts */
-			eth_irq ( ENABLE );
-			undi_isr->FuncFlag = PXENV_UNDI_ISR_OUT_DONE;
+			/* No break - fall through to IN_GET_NEXT */
 		}
-		break;
 	case PXENV_UNDI_ISR_IN_GET_NEXT :
 		/* We only ever return one frame at a time */
 		DBG ( " GET_NEXT DONE" );
 		/* Re-enable interrupts */
 		eth_irq ( ENABLE );
+		/* Force an interrupt if there's a packet still
+		 * waiting, since we only handle one packet per
+		 * interrupt.
+		 */
+		if ( eth_poll ( 0 ) ) {
+			DBG ( " (RETRIGGER)" );
+			eth_irq ( FORCE );
+		}
 		undi_isr->FuncFlag = PXENV_UNDI_ISR_OUT_DONE;
 		break;
 	default :
@@ -828,8 +839,12 @@ PXENV_EXIT_t pxenv_tftp_read ( t_PXENV_TFTP_READ *tftp_read ) {
  */
 
 int pxe_tftp_read_block ( unsigned char *data, unsigned int block __unused, unsigned int len, int eof ) {
-	if ( pxe_stack->readfile.offset + len >= pxe_stack->readfile.bufferlen ) return -1;
-	memcpy ( pxe_stack->readfile.buffer + pxe_stack->readfile.offset, data, len );
+	if ( pxe_stack->readfile.buffer ) {
+		if ( pxe_stack->readfile.offset + len >=
+		     pxe_stack->readfile.bufferlen ) return -1;
+		memcpy ( pxe_stack->readfile.buffer +
+			 pxe_stack->readfile.offset, data, len );
+	}
 	pxe_stack->readfile.offset += len;
 	return eof ? 0 : 1;
 }
@@ -862,13 +877,29 @@ PXENV_EXIT_t pxenv_tftp_read_file ( t_PXENV_TFTP_READ_FILE *tftp_read_file ) {
 
 /* PXENV_TFTP_GET_FSIZE
  *
- * Status: stub
+ * Status: working, though ugly (we actually read the whole file,
+ * because it's too ugly to make Etherboot request the tsize option
+ * and hand it to us).
  */
 PXENV_EXIT_t pxenv_tftp_get_fsize ( t_PXENV_TFTP_GET_FSIZE *tftp_get_fsize ) {
+	int rc;
+
 	DBG ( "PXENV_TFTP_GET_FSIZE" );
-	/* ENSURE_READY ( tftp_get_fsize ); */
-	tftp_get_fsize->Status = PXENV_STATUS_UNSUPPORTED;
-	return PXENV_EXIT_FAILURE;
+	ENSURE_READY ( tftp_get_fsize );
+
+	pxe_stack->readfile.buffer = NULL;
+	pxe_stack->readfile.bufferlen = 0;
+	pxe_stack->readfile.offset = 0;
+
+	rc = tftp ( tftp_get_fsize->FileName, pxe_tftp_read_block );
+	if ( rc ) {
+		tftp_get_fsize->FileSize = 0;
+		tftp_get_fsize->Status = PXENV_STATUS_FAILURE;
+		return PXENV_EXIT_FAILURE;
+	}
+	tftp_get_fsize->FileSize = pxe_stack->readfile.offset;
+	tftp_get_fsize->Status = PXENV_STATUS_SUCCESS;
+	return PXENV_EXIT_SUCCESS;
 }
 
 /* PXENV_UDP_OPEN
@@ -1152,7 +1183,7 @@ PXENV_EXIT_t pxenv_restart_tftp ( t_PXENV_RESTART_TFTP *restart_tftp ) {
 
 /* PXENV_START_BASE
  *
- * Status: stub
+ * Status: won't implement (requires major structural changes)
  */
 PXENV_EXIT_t pxenv_start_base ( t_PXENV_START_BASE *start_base ) {
 	DBG ( "PXENV_START_BASE" );
@@ -1163,18 +1194,23 @@ PXENV_EXIT_t pxenv_start_base ( t_PXENV_START_BASE *start_base ) {
 
 /* PXENV_STOP_BASE
  *
- * Status: stub
+ * Status: working
  */
 PXENV_EXIT_t pxenv_stop_base ( t_PXENV_STOP_BASE *stop_base ) {
 	DBG ( "PXENV_STOP_BASE" );
-	/* ENSURE_READY ( stop_base ); */
+
+	/* The only time we will be called is when the NBP is trying
+	 * to shut down the PXE stack.  There's nothing we need to do
+	 * in this call.
+	 */
+
 	stop_base->Status = PXENV_STATUS_SUCCESS;
 	return PXENV_EXIT_SUCCESS;
 }
 
 /* PXENV_UNDI_LOADER
  *
- * Status: stub
+ * Status: working
  *
  * NOTE: This is not a genuine PXE API call; the loader has a separate
  * entry point.  However, to simplify the mapping of the PXE API to
